"""Plays files from the local filesystem."""

import functools
import logging
import time
import urllib.parse
from http.server import HTTPServer
from ipaddress import IPv4Address, IPv4Network
from os import chdir, path
from socketserver import ThreadingMixIn
from sys import version_info as pyversion
from threading import Thread

import ifaddr
from RangeHTTPServer import RangeRequestHandler

from soco_cli.utils import (
    remember_event_sub,
    error_report,
    event_unsubscribe,
    forget_event_sub,
    set_speaker_playing_local_file,
)

# The HTTP server port range to use
PORT_START = 54000
PORT_END = 54099

SUPPORTED_TYPES = ["MP3", "M4A", "MP4", "FLAC", "OGG", "WMA", "WAV"]

# Need to know whether this is Python >= 3.7
PY37PLUS = bool((pyversion.major == 3 and pyversion.minor >= 7) or pyversion.major > 3)


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in separate threads.

    Use the MixIn approach instead of the core ThreadingHTTPServer
    class for backwards compatibility with Python 3.5+
    """


class MyHTTPHandler(RangeRequestHandler):
    # Handle the change to the SimpleHTTPRequestHandler __init__() in Python 3.7+
    if PY37PLUS:

        def __init__(self, *args, filename=None, speaker_ips=None, **kwargs):
            self.filename = filename
            self.speaker_ips = speaker_ips
            super().__init__(*args, **kwargs)

    else:

        def __init__(
            self, *args, filename=None, speaker_ips=None, directory="", **kwargs
        ):
            self.filename = filename
            self.speaker_ips = speaker_ips
            try:
                chdir(directory)
            except:
                pass
            super().__init__(*args, **kwargs)

    def do_GET(self):
        logging.info("Get request received by HTTP server")

        # Only serve the specific file requested on the command line,
        # and only to Sonos speakers in the Sonos system
        error = False
        if self.path.replace("/", "") != self.filename:
            logging.info("Access to file '{}' forbidden".format(self.path))
            error = True
        if self.client_address[0] not in self.speaker_ips:
            logging.info("Access from IP '{}' forbidden".format(self.client_address[0]))
            error = True
        if error:
            RangeRequestHandler.send_error(
                self, code=403, message="SoCo-CLI HTTP Server: Access forbidden"
            )
            return

        # Forward the GET request
        try:
            super().do_GET()
        except Exception as e:
            # It's normal to hit some exceptions with Sonos.
            logging.info("Exception ignored: {}".format(e))

    def log_message(self, format, *args):
        # Suppress HTTP logging
        return


def http_server(server_ip, directory, filename, speaker_ips):
    # Set the directory from which to serve files, in the handler
    # Set the specific filename and client IP that are authorised
    handler = functools.partial(
        MyHTTPHandler, filename=filename, speaker_ips=speaker_ips, directory=directory
    )

    # For possible future use: set up MIME types
    # MyHTTPHandler.extensions_map[".m4a"] = "audio/x-m4a"
    # MyHTTPHandler.extensions_map[".aac"] = "audio/aac"

    # Find an available port by trying ports in sequence
    for port in range(PORT_START, PORT_END + 1):
        try:
            httpd = ThreadedHTTPServer((server_ip, port), handler)
            logging.info("Using {}:{} for web server".format(server_ip, port))
            httpd_thread = Thread(target=httpd.serve_forever, daemon=True)
            httpd_thread.start()
            logging.info("Web server started")
            return httpd
        except OSError:
            # Assume this means that the port is in use
            continue
    return None


def get_server_ip(speaker):
    # Get a suitable IP address to use as a server address for Sonos
    # on this host
    adapters = ifaddr.get_adapters()
    for adapter in adapters:
        for ip in adapter.ips:
            if ip.is_IPv4:
                network = IPv4Network(
                    ip.ip + "/" + str(ip.network_prefix), strict=False
                )
                if IPv4Address(speaker.ip_address) in network:
                    return ip.ip
    return None


def wait_until_stopped(speaker, uri):
    playing_states = ["PLAYING", "TRANSITIONING", "PAUSED_PLAYBACK"]
    try:
        sub = speaker.avTransport.subscribe(auto_renew=True)
        remember_event_sub(sub)
    except Exception as e:
        error_report("Exception {}".format(e))
        return

    while True:
        try:
            event = sub.events.get(timeout=1.0)
            state = event.variables["transport_state"]
            logging.info("Event received: playback state = '{}'".format(state))

            if state not in playing_states:
                logging.info(
                    "Speaker '{}' in state '{}'".format(
                        speaker.player_name, event.variables["transport_state"]
                    )
                )
                break

            # Check that the expected URI is still playing
            try:
                current_uri = event.variables["current_track_meta_data"].get_uri()
            except:
                # Can only call get_uri() on certain datatypes
                current_uri = ""
            if current_uri != uri:
                logging.info("Playback URI changed: exit event wait loop")
                break

        except:
            pass

    event_unsubscribe(sub)
    forget_event_sub(sub)
    return


def is_supported_type(filename):
    file_upper = filename.upper()
    for file_type in SUPPORTED_TYPES:
        if file_upper.endswith("." + file_type):
            # Supported file type
            return True
    return False


def play_local_file(speaker, pathname):
    # speaker is a SoCo instance
    # pathname is the local file to be played

    if not path.exists(pathname):
        error_report("File '{}' not found".format(pathname))
        return False

    directory, filename = path.split(pathname)

    if not is_supported_type(filename):
        error_report(
            "Unsupported file type; must be one of: {}".format(SUPPORTED_TYPES)
        )
        return False

    # Make filename compatible with URL naming
    url_filename = urllib.parse.quote(filename)

    server_ip = get_server_ip(speaker)
    if not server_ip:
        error_report("Can't determine an IP address for web server")
        return False
    logging.info("Using server IP address: {}".format(server_ip))

    # Start the webserver (runs in a daemon thread)
    speaker_ips = []
    for zone in speaker.all_zones:
        speaker_ips.append(zone.ip_address)
    httpd = http_server(server_ip, directory, url_filename, speaker_ips)
    if not httpd:
        error_report("Cannot create HTTP server")
        return False

    # This ensures that other running invocations of 'play_file'
    # receive their stop events, and terminate.
    logging.info("Stopping speaker '{}'".format(speaker.player_name))
    speaker.stop()

    # Assemble the URI
    uri = "http://" + server_ip + ":" + str(httpd.server_port) + "/" + url_filename
    logging.info("Playing file '{}' from directory '{}'".format(filename, directory))
    logging.info("Playback URI: {}".format(uri))

    logging.info("Send URI to '{}' for playback".format(speaker.player_name))
    speaker.play_uri(uri)

    logging.info("Setting flag to stop playback on CTRL-C")
    set_speaker_playing_local_file(speaker)

    logging.info("Waiting 3s for playback to start")
    time.sleep(3.0)
    logging.info("Waiting for playback to stop")
    wait_until_stopped(speaker, uri)
    logging.info("Playback stopped ... terminating web server")
    httpd.shutdown()
    logging.info("Web server terminated")

    set_speaker_playing_local_file(None)

    return True



"""Manages aliases for use with the interactive shell."""

import logging
import pickle
from os import mkdir, path

CONFIG_DIR = path.join(path.expanduser("~"), ".soco-cli")
ALIAS_FILE = path.join(CONFIG_DIR, "aliases.pickle")


class AliasManager:
    def __init__(self):
        self._aliases = {}

    def create_alias(self, alias_name, alias_actions):
        alias_name = alias_name.strip()
        if alias_actions:
            alias_actions = alias_actions.strip()
        if alias_actions in [None, ""]:
            return self.remove_alias(alias_name)
        if self._aliases.get(alias_name, None):
            new = False
        else:
            new = True
        logging.info("Adding alias '{}', new = {}".format(alias_name, new))
        self._aliases[alias_name] = alias_actions
        return True, new

    def action(self, alias_name):
        return self._aliases.get(alias_name, None)

    def remove_alias(self, alias_name):
        alias_name = alias_name.strip()
        try:
            del self._aliases[alias_name]
            logging.info("Removing alias '{}'".format(alias_name))
            return True
        except KeyError:
            logging.info("Alias '{}' not found".format(alias_name))
            return False

    def alias_names(self):
        return list(self._aliases.keys())

    def save_aliases(self):
        if not path.exists(CONFIG_DIR):
            try:
                logging.info("Creating directory '{}'".format(CONFIG_DIR))
                mkdir(CONFIG_DIR)
            except:
                pass
        with open(ALIAS_FILE, "wb") as f:
            logging.info("Saving aliases")
            pickle.dump(self._aliases, f)

    def load_aliases(self):
        logging.info("Reading aliases")
        try:
            with open(ALIAS_FILE, "rb") as f:
                self._aliases = pickle.load(f)
        except:
            logging.info("Failed to read aliases from file")

    def print_aliases(self):
        if len(self._aliases) == 0:
            print("No current aliases")
            return
        print()
        print(self._aliases_to_text())

    def save_aliases_to_file(self, filename):
        try:
            with open(filename, "w") as f:
                f.write("# Soco-CLI Aliases File\n")
                f.write(self._aliases_to_text(raw=True))
                return True
        except:
            return False

    def load_aliases_from_file(self, filename):
        try:
            with open(filename, "r") as f:
                line = f.readline()
                while line != "":
                    if not line.startswith("#") and line != "\n":
                        if line.count("=") != 1:
                            print("Malformed alias ... ignored")
                            print(line, end="")
                        else:
                            alias = line.split("=")
                            self.create_alias(alias[0], alias[1])
                    line = f.readline()
            self.save_aliases()
            return True
        except:
            return False

    def _aliases_to_text(self, raw=False):
        output = ""
        if not raw:
            max_alias = len(max(self._aliases.keys(), key=len))
        for alias_name in sorted(self._aliases.keys()):
            if raw:
                output = output + alias_name + " = " + self._aliases[alias_name] + "\n"
            else:
                output = (
                    output
                    + "  "
                    + alias_name.ljust(max_alias)
                    + " = "
                    + self._aliases[alias_name]
                    + "\n"
                )
        return output



"""SoCo-CLI is a command line control interface for Sonos systems.

It is a simplified wrapper around the SoCo python library, as well as providing
an extensive range of additional features.

It can be used as a command line program, as an interactive command shell, and
in other programs via its simple API. It can also run as a simple HTTP API
server, to control Sonos via HTTP requests.

For more information, please see: https://github.com/avantrec/soco-cli
"""

__version__ = "0.4.14"



"""Plays a list of files from the local filesystem, with interactive options."""

import logging
import os
import sys
from multiprocessing import Process
from os import chdir, name, path
from pathlib import Path
from random import choice, sample

from soco import SoCo

from soco_cli.m3u_parser import parse_m3u
from soco_cli.play_local_file import is_supported_type, play_local_file
from soco_cli.utils import error_report

# def wait_for_keypress():
#     # Wait for a key press on the console and return it
#     result = None
#
#     if name == "nt":  # Windows
#         import msvcrt
#
#         result = msvcrt.getch().decode()
#     else:  # Linux & macOS
#         import termios
#
#         fd = sys.stdin.fileno()
#         oldterm = termios.tcgetattr(fd)
#         newattr = termios.tcgetattr(fd)
#         newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
#         termios.tcsetattr(fd, termios.TCSANOW, newattr)
#         try:
#             result = sys.stdin.read(1)
#         except IOError:
#             pass
#         finally:
#             termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
#     return result


def interaction_manager(speaker_ip):
    sys.stdin = open(0)
    speaker = SoCo(speaker_ip)
    while True:
        try:
            # keypress = wait_for_keypress()
            keypress = input("")[0]
        except:
            keypress = None
        if keypress in ["N", "n"]:
            action = "NEXT"
            print("Next track ...")
            speaker.stop()
            logging.info(
                "Interactive mode: key = '{}', action = '{}'".format(keypress, action)
            )
        if keypress in ["P", "p"]:
            action = "PAUSE"
            print("Pause playback ...")
            try:
                speaker.pause()
            except Exception as e:
                logging.info("Exception ignored: {}".format(e))
            logging.info(
                "Interactive mode: key = '{}', action = '{}'".format(keypress, action)
            )
        if keypress in ["R", "r"]:
            action = "RESUME"
            print("Resume playback ...")
            try:
                speaker.play()
            except Exception as e:
                logging.info("Exception ignored: {}".format(e))
            logging.info(
                "Interactive mode: key = '{}', action = '{}'".format(keypress, action)
            )
        # Windows captures CTRL-C key-presses, so we handle them directly here
        if name == "nt" and keypress == "\x03":
            logging.info(
                "Windows CTRL-C: Stopping speaker '{}' and exiting".format(
                    speaker.player_name
                )
            )
            speaker.stop()
            os._exit(0)


def play_m3u_file(speaker, m3u_file, options=""):
    """Play a M3U or M3U8 file"""
    options = options.lower()

    # Check for invalid options
    invalid = set(options) - set("psri")
    if invalid:
        error_report("Invalid option(s) '{}' supplied".format(invalid))
        return False

    if not path.exists(m3u_file):
        error_report("File '{}' not found".format(m3u_file))
        return False

    # if not (m3u_file.lower().endswith(".m3u") or m3u_file.lower().endswith(".m3u8")):
    #     error_and_exit(
    #         "Filename '{}' does not end in '.m3u' or '.m3u8'".format(m3u_file)
    #     )
    #     return False

    logging.info("Parsing file contents'{}'".format(m3u_file))
    tracks = parse_m3u(m3u_file)
    if not tracks:
        error_report("No tracks found in '{}'".format(m3u_file))

    logging.info("Found {} tracks".format(len(tracks)))

    if "r" in options:
        # Choose a single random track
        track = choice(tracks)
        tracks = [track]
        logging.info("Choosing random track: {}".format(track.path))

    elif "s" in options:
        logging.info("Shuffling playlist")
        # For some reason, 'shuffle(tracks)' does not work
        tracks = sample(tracks, len(tracks))

    directory, _ = path.split(m3u_file)
    if directory != "":
        chdir(directory)

    # Interactive mode
    keypress_process = None
    if "i" in options:
        print("Interactive mode actions: (N)ext, (P)ause, (R)esume + RETURN")
        try:
            logging.info("Interactive mode ... starting keypress process")
            keypress_process = Process(
                target=interaction_manager, args=(speaker.ip_address,), daemon=True
            )
            keypress_process.start()
            logging.info("Process PID {} created".format(keypress_process.pid))
        except Exception as e:
            logging.info("Exception ignored: {}".format(e))
            keypress_process = None

    zero_pad = len(str(len(tracks)))
    for index, track in enumerate(tracks):
        abs_filename = str(Path(track.path).absolute())
        logging.info("Convert '{}' to '{}'".format(track.path, abs_filename))

        if not path.exists(abs_filename):
            print("Error: file not found:", abs_filename)
            continue

        if not is_supported_type(abs_filename):
            print("Error: unsupported file type:", abs_filename)
            continue

        if "p" in options:
            print(
                "Playing {} of {}:".format(str(index + 1).zfill(zero_pad), len(tracks)),
                abs_filename,
            )

        play_local_file(speaker, abs_filename)

    if keypress_process:
        keypress_process.terminate()

    if "p" in options:
        print("End of playlist")

    return True



"""Implements an HTTP API server for SoCo-CLI commands."""

from sys import version_info

if version_info.major == 3 and version_info.minor < 6:
    print("HTTP API Server requires Python 3.6 or above")
    exit(1)

import argparse

import uvicorn
from fastapi import FastAPI

from soco_cli.__init__ import __version__ as version
from soco_cli.api import get_all_speaker_names
from soco_cli.api import get_soco_object as get_speaker
from soco_cli.api import rescan_speakers
from soco_cli.api import run_command as sc_run
from soco_cli.utils import version as print_version

sc_app = FastAPI()

# Globals
USE_LOCAL = False
PORT = 8000
INFO = "SoCo-CLI HTTP API Server v" + version
PREFIX = "SoCo-CLI: "


def command_core(speaker, action, *args, use_local=False):
    device, error_msg = get_speaker(speaker, use_local_speaker_list=use_local)
    if device:
        speaker = device.player_name
        exit_code, result, error_msg = sc_run(
            device, action, *args, use_local_speaker_list=use_local
        )
    else:
        exit_code = 1
        result = ""

    # Quote speaker names & arguments containing spaces, for neatness
    if " " in speaker:
        quoted_speaker = '"' + speaker + '"'
    else:
        quoted_speaker = speaker
    new_args = []
    for i in range(len(args)):
        if " " in args[i]:
            new_args.append('"' + args[i] + '"')
        else:
            new_args.append(args[i])

    # Print the equivalent 'sonos' command and exit code
    if len(new_args) != 0:
        arguments = " ".join(new_args).rstrip()
        print(
            PREFIX
            + "Command = 'sonos {} {} {}', ".format(quoted_speaker, action, arguments),
            end="",
        )
    else:
        print(
            PREFIX + "Command = 'sonos {} {}', ".format(quoted_speaker, action), end=""
        )
    if exit_code == 0:
        print("exit code = {}".format(exit_code))
    else:
        print("exit code = {} [{}]".format(exit_code, error_msg))

    return {
        "speaker": speaker,
        "action": action,
        "args": args,
        "exit_code": exit_code,
        "result": result,
        "error_msg": error_msg,
    }


@sc_app.get("/")
async def root():
    return {"info": INFO}


@sc_app.get("/rediscover")
async def rediscover():
    rescan_speakers(timeout=2.0)
    speakers = get_all_speaker_names()
    print(PREFIX + "Speakers (re)discovered: {}".format(speakers))
    return {"speakers_discovered": speakers}


@sc_app.get("/{speaker}/{action}")
async def action_0(speaker: str, action: str):
    return command_core(speaker, action, use_local=USE_LOCAL)


@sc_app.get("/{speaker}/{action}/{arg_1}")
async def action_1(speaker: str, action: str, arg_1: str):
    return command_core(speaker, action, arg_1, use_local=USE_LOCAL)


@sc_app.get("/{speaker}/{action}/{arg_1}/{arg_2}")
async def action_2(speaker: str, action: str, arg_1: str, arg_2: str):
    return command_core(speaker, action, arg_1, arg_2, use_local=USE_LOCAL)


@sc_app.get("/{speaker}/{action}/{arg_1}/{arg_2}/{arg_3}")
async def action_3(speaker: str, action: str, arg_1: str, arg_2: str, arg_3: str):
    return command_core(speaker, action, arg_1, arg_2, arg_3, use_local=USE_LOCAL)


def args_processor():
    parser = argparse.ArgumentParser(
        prog="sonos-http-api-server",
        usage="%(prog)s",
        description=INFO,
    )
    parser.add_argument(
        "--port",
        "-p",
        type=int,
        help="The port on which to listen",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        default=False,
        help="Print the SoCo-CLI and SoCo versions, and exit",
    )

    args = parser.parse_args()

    if args.version:
        print_version()
        exit(0)

    global PORT
    if args.port is not None:
        PORT = args.port


def main():
    args_processor()
    print(PREFIX + "Starting " + INFO)

    try:
        # Pre-load speaker cache
        print(PREFIX + "Finding speakers ... ", end="", flush=True)
        try:
            # This forces speaker discovery
            # For some reason, using 'get_all_speakers()' provokes Uvicorn errors
            get_speaker("")
            print(get_all_speaker_names())
        except:
            print("discovery failed, will retry on first request")

        # Start the server
        uvicorn.run(sc_app, host="0.0.0.0", use_colors=False, port=PORT)
        print(PREFIX + INFO + " stopped")
        exit(0)

    except Exception as error:
        print("Error: {}".format(error))
        exit(1)


if __name__ == "__main__":
    main()



"""Parse sequential command lines, using ':' as a command separator."""


class CLIParser:
    def __init__(self):
        self._args = None
        self._sequences = None
        self._separator = ":"

    def parse(self, args):

        self._args = args

        sequence = []  # A single command sequence
        sequences = []  # A list of command sequences
        for arg in args:
            # if len(arg) > 1 and self._separator in arg:
            #     # Catch special cases of colon use: HH:MM(:SS) time formats,
            #     # and URLs
            #     if not (
            #         sequence
            #         and sequence[-1]
            #         in [
            #             "wait",
            #             "wait_for",
            #             "wait_until",
            #             "seek",
            #             "seek_to",
            #             "seek_forward",
            #             "sf",
            #             "seek_back",
            #             "sb",
            #             "sleep",
            #             "sleep_timer",
            #             "sleep_at",
            #             "wait_stopped_for",
            #             "loop_for",
            #             "loop_until",
            #         ]
            #         or ":/" in arg
            #     ):
            #         error_and_exit(
            #             "Spaces are required each side of the ':' command separator"
            #         )
            if arg != self._separator:
                sequence.append(arg)
            else:
                sequences.append(sequence)
                sequence = []
        if sequence:
            sequences.append(sequence)

        self._sequences = sequences

    def get_sequences(self):
        return self._sequences



"""SoCo-CLI interactive mode handler."""

import logging
import subprocess
import sys

# Readline is only available on Unix
try:
    import readline

    RL = True
except ImportError:

    RL = False

from os import chdir
from shlex import split as shlex_split

from soco_cli.action_processor import get_actions, list_actions
from soco_cli.aliases import AliasManager
from soco_cli.api import get_soco_object, run_command
from soco_cli.check_for_update import print_update_status
from soco_cli.cmd_parser import CLIParser
from soco_cli.keystroke_capture import get_keystroke
from soco_cli.utils import (
    RewindableList,
    docs,
    get_readline_history,
    get_speaker,
    local_speaker_list,
    save_readline_history,
    set_interactive,
    set_suspend_sighandling,
    speaker_cache,
    version,
)

from .wait_actions import process_wait

# Alias Manager
am = AliasManager()

# The following actions are run in a subprocess, to allow them to be terminated
# without dropping out of the interactive shell.
ACTIONS_TO_EXEC = [
    "track_follow",
    "tf",
    "track_follow_compact",
    "tfc",
    "wait_stop",
    "wait_start",
    "wait_stopped_for",
    "wsf",
    "wait_stop_not_pause",
    "wsnp",
    "wait_stopped_for_not_pause",
    "wsfnp",
    "wait_end_track",
]


ACTIONS_TO_EXEC_NO_SPEAKER = [
    "wait",
    "wait_until",
    "wait_for",
]


def interactive_loop(
    speaker_name, use_local_speaker_list=False, no_env=False, single_keystroke=False
):
    """
    The main interactive loop for gathering and processing interactive commands.

    Args:
        speaker_name (str): The name of the speaker supplied if supplied on the command
            line.
        use_local_speaker_list (bool): Whether to use cached discovery.
        no_env (bool): Whether to ignore environment variables.
        single_keystroke (bool): Whether to start in single keystroke mode.
    """

    speaker = None
    saved_speaker = None
    pushed = False
    temp_active_speaker = False

    # Is the speaker name set on the command line?
    # Note: ignores SPKR set as part of the environment
    if speaker_name:
        try:
            speaker, error_msg = get_soco_object(
                speaker_name, use_local_speaker_list=use_local_speaker_list
            )
            if not speaker:
                print("Speaker '{}' not found [{}]".format(speaker_name, error_msg))
                speaker_name = None
            else:
                speaker_name = speaker.player_name
        except Exception as e:
            print("Error finding speaker '{}': {}".format(speaker_name, e))
            speaker_name = None

    print("\nEntering SoCo-CLI interactive shell.")
    if single_keystroke:
        print("Single Keystroke Mode ... 'x' to exit.\n")
    else:
        print("Type 'help' for available shell commands.\n")
    if not RL:
        print("Note: Autocompletion not currently available on Windows.\n")

    set_interactive()
    am.load_aliases()

    if RL:
        _set_actions_and_commands_list(use_local_speaker_list=use_local_speaker_list)
        readline.parse_and_bind("tab: complete")
        readline.set_completer(_completer)
        readline.set_completer_delims(" ")
        _get_readline_history()

    root_prompt = "Sonos"

    # Input loop
    while True:
        # Catch all exceptions raised in the input loop
        try:
            if speaker_name and speaker:
                prompt = (root_prompt + " [{}] > ").format(speaker_name)
            else:
                prompt = root_prompt + " [] > "

            if single_keystroke:
                prompt = prompt.replace(">", ">>")
                print(prompt, flush=True, end="")
                command_line = get_keystroke()
                print(command_line)
                # Catch exit, including Windows CTRL-C
                if command_line in ["x", "\x03"]:
                    logging.info("Exit from single keystroke mode")
                    single_keystroke = False
                    continue
            else:
                command_line = input(prompt)

            if command_line == "":
                continue

            # Parse multiple action sequences
            cli_parser = CLIParser()
            try:
                cli_parser.parse(shlex_split(command_line))
            except ValueError as error:
                print("Error: {}".format(error))
                continue

            # Loop through action sequences
            command_sequences = RewindableList(cli_parser.get_sequences())
            logging.info("Command sequences = {}".format(command_sequences))

            # The command_sequence list can change, so we use pop_next() until the
            # list is exhausted
            while True:
                try:
                    command = command_sequences.pop_next()
                    logging.info("Current sequence = {}".format(command))
                except IndexError:
                    break

                if len(command) == 0:
                    continue

                # Is this an alias?
                if command[0] in am.alias_names():
                    ap = AliasProcessor()
                    index = command_sequences.index()
                    ap.process(command, am, command_sequences)
                    command_sequences.rewind_to(index)
                    continue

                if command[0] == "0":
                    # Unset the active speaker
                    logging.info("Unset active speaker")
                    speaker_name = None
                    speaker = None
                    continue

                command_lower = command[0].lower()

                if command_lower.startswith("exit"):
                    logging.info("Exiting interactive mode")
                    _save_readline_history()
                    return True

                if command_lower in ["help", "?"]:
                    _interactive_help()
                    continue

                if command_lower == "actions":
                    _show_actions()
                    continue

                if command_lower in ["single-keystroke", "sk"]:
                    print("Single keystroke mode ... 'x' to exit")
                    single_keystroke = True
                    continue

                if command_lower == "speakers":
                    _print_speaker_list(use_local_speaker_list=use_local_speaker_list)
                    continue

                # if command_lower in ["wait", "wait_until", "wait_for"]:
                #     logging.info("Processing a 'wait' action")
                #     process_wait(command)
                #     continue

                if command_lower in ["version"]:
                    print()
                    version()
                    print()
                    continue

                if command_lower in ["docs"]:
                    docs()
                    continue

                if command_lower in ["check_for_update"]:
                    print_update_status()
                    continue

                # Is the input a number in the range of speaker numbers?
                try:
                    speaker_number = int(command_lower)
                    limit = len(
                        _get_speaker_names(
                            use_local_speaker_list=use_local_speaker_list
                        )
                    )
                    if 1 <= speaker_number <= limit:
                        logging.info(
                            "Setting to active speaker no. {} ".format(speaker_number)
                        )
                        speaker_name = _get_speaker_names(
                            use_local_speaker_list=use_local_speaker_list
                        )[speaker_number - 1]
                        speaker = get_speaker(speaker_name, use_local_speaker_list)
                        logging.info("{} : {}".format(speaker, speaker_name))
                    else:
                        print(
                            "Error: Speaker number is out of range (0 to {})".format(
                                limit
                            )
                        )
                    continue
                except ValueError:
                    pass

                # if command_lower == "rediscover":
                #     if use_local_speaker_list:
                #         print("Using cached speaker list: no discovery performed")
                #     else:
                #         speaker_cache().discover(reset=True)
                #     continue

                if command_lower == "rescan":
                    _rescan(use_local_speaker_list=use_local_speaker_list)
                    continue

                if command_lower == "rescan_max":
                    _rescan(
                        use_local_speaker_list=use_local_speaker_list, max_scan=True
                    )
                    continue

                if command_lower == "exec":
                    if len(command) > 1:
                        _exec(command[1:])
                        # print()
                    continue

                if command_lower == "cd":
                    if len(command) > 1:
                        try:
                            logging.info("Attempting to cd to: '{}'".format(command[1]))
                            chdir(command[1])
                        except Exception as e:
                            print(e)
                    continue

                if command_lower == "push":
                    if pushed is True:
                        logging.info("Active speaker already pushed ... ignored")
                        continue
                    if speaker:
                        pushed = True
                        logging.info(
                            "Pushing current active speaker: {}".format(
                                speaker.player_name
                            )
                        )
                    else:
                        pushed = False
                        logging.info("No active speaker to push")
                    saved_speaker = speaker
                    speaker_name = None
                    speaker = None
                    continue

                if command_lower == "pop":
                    if pushed is False:
                        logging.info("No active speaker state to pop ... ignored")
                        continue
                    logging.info("Popping the saved speaker state")
                    if saved_speaker:
                        speaker = saved_speaker
                        speaker_name = speaker.player_name
                        logging.info("Saved speaker = '{}'".format(speaker_name))
                        saved_speaker = None
                    elif pushed:
                        saved_speaker = None
                        speaker = None
                        speaker_name = None
                    else:
                        logging.info("No saved speaker")
                    pushed = False
                    continue

                # Alias creation, update, and deletion
                if command_lower == "alias":
                    if len(command) == 1:
                        am.print_aliases()
                        continue
                    # Remove 'alias'
                    command.pop(0)
                    alias_name = command.pop(0)
                    if alias_name == "alias":
                        print("Not permitted: cannot create alias for 'alias'")
                        continue
                    if len(command) == 0:
                        if am.create_alias(alias_name, None):
                            print("Alias '{}' removed".format(alias_name))
                        else:
                            print("Alias '{}' not found".format(alias_name))
                    else:
                        # Have to collect the remaining sequences: they're all
                        # part of the alias. Reconstruction required.
                        # Multi-word parameters need quotes reinstated
                        _restore_quotes(command)
                        actions = [" ".join(command)]
                        logging.info("Action = '{}'".format(command))
                        while True:
                            try:
                                command = command_sequences.pop_next()
                                _restore_quotes(command)
                                actions.append(" ".join(command))
                                logging.info("Action = '{}'".format(command))
                            except IndexError:
                                break
                        action = " : ".join(actions)
                        logging.info("Action sequence = '{}'".format(action))
                        _, new = am.create_alias(alias_name, action)
                        if new:
                            print("Alias '{}' created".format(alias_name))
                        else:
                            print("Alias '{}' updated".format(alias_name))
                    am.save_aliases()
                    _set_actions_and_commands_list(
                        use_local_speaker_list=use_local_speaker_list
                    )
                    continue

                # Command processing
                try:
                    args = command

                    # Setting a speaker to operate on?
                    try:
                        if args[0] == "set":
                            new_speaker_name = args[1]
                            new_speaker = get_speaker(
                                new_speaker_name, use_local_speaker_list
                            )
                            if not new_speaker:
                                print(
                                    "Error: Speaker '{}' not found".format(
                                        new_speaker_name
                                    )
                                )
                            else:
                                logging.info(
                                    "Set new active speaker: '{}'".format(
                                        new_speaker_name
                                    )
                                )
                                speaker = new_speaker
                                speaker_name = speaker.player_name
                            continue
                    except IndexError:
                        # No speaker name given
                        logging.info("Unset active speaker ('set' with no arguments)")
                        speaker_name = None
                        speaker = None
                        continue

                    if not speaker_name:
                        logging.info(
                            "Treating first parameter '{}' as speaker name".format(
                                args[0]
                            )
                        )
                        name = args.pop(0)
                        speaker = get_speaker(name, use_local_speaker_list)
                        if not speaker:
                            print(
                                "Error: Speaker '{}' not found; should an active speaker be set?".format(
                                    name
                                )
                            )
                            continue
                        if len(args) == 0:
                            print(
                                "Error: no action or arguments supplied for speaker '{}'".format(
                                    speaker.player_name
                                )
                            )
                            continue

                        # Temporarily establish an active speaker
                        temp_active_speaker = True
                        speaker_name = speaker.player_name
                        logging.info(
                            "Temporarily establish active speaker: '{}'".format(
                                speaker_name
                            )
                        )
                        # Replace the command sequence without the speaker name,
                        # for processing next time round the loop
                        command_sequences.insert(command_sequences.index(), args)
                        logging.info("Reinserting command sequence: {}".format(args))
                        continue

                    action = args.pop(0).lower()
                    logging.info("Action = '{}'; args = {}".format(action, args))
                    # Commands often requiring CTRL-C to exit are run in a subprocess
                    if (
                        action in ACTIONS_TO_EXEC
                        or action in ACTIONS_TO_EXEC_NO_SPEAKER
                    ):
                        _exec_action(speaker.ip_address, action, args)
                    else:
                        exit_code, output, error_msg = run_command(
                            speaker,
                            action,
                            *args,
                            use_local_speaker_list=use_local_speaker_list,
                        )
                        if exit_code:
                            if error_msg != "":
                                print(error_msg)
                        else:
                            if output != "":
                                print(output)
                            if action == "rename":
                                speaker_name = speaker.get_speaker_info(refresh=True)[
                                    "zone_name"
                                ]
                                _set_actions_and_commands_list(
                                    use_local_speaker_list=use_local_speaker_list
                                )
                    if temp_active_speaker:
                        logging.info(
                            "Unsetting temporary active speaker: '{}'".format(
                                speaker_name
                            )
                        )
                        temp_active_speaker = False
                        speaker = None
                        speaker_name = None
                except:
                    print("Error: Invalid command")

        # Catch all exceptions in the interactive loop
        except Exception as e:
            print("Error: {}".format(e))


COMMANDS = [
    "actions",
    "alias ",
    "cd",
    "check_for_update",
    "docs",
    "exec",
    "exit",
    "help",
    "pop",
    "push",
    "rescan",
    "rescan_max",
    "set ",
    "single-keystroke",
    "sk",
    "speakers",
    "version",
]


def _completer(text, context):
    """Auto-complete commands using TAB"""
    matches = [cmd for cmd in _get_actions_and_commands() if cmd.startswith(text)]
    return matches[context]


def _show_actions():
    print()
    print("Complete list of SoCo-CLI actions:")
    print("==================================")
    print()
    list_actions(
        include_loop_actions=False,
        include_wait_actions=True,
        include_track_follow_actions=True,
    )
    print()


ACTIONS_LIST = []


def _set_actions_and_commands_list(use_local_speaker_list=False):
    logging.info("Rebuilding commands/action list")
    global ACTIONS_LIST
    ACTIONS_LIST = (
        [
            action + " "
            for action in get_actions(
                include_loop_actions=False,
                include_wait_actions=True,
                include_track_follow_actions=True,
            )
            + _get_speaker_names(use_local_speaker_list=use_local_speaker_list)
        ]
        + COMMANDS
        + am.alias_names()
    )


def _get_actions_and_commands():
    return ACTIONS_LIST


def _interactive_help():
    print(HELP_TEXT)


HELP_TEXT = """
This is SoCo-CLI interactive mode. Interactive commands are as follows:

    '1', ...     :  Set the active speaker. Use the numbers shown by the
                    'speakers' command. E.g., to set to speaker number 4
                    in the list, just type '4'.
                    '0' will unset the active speaker.
    'actions'    :  Show the complete list of SoCo-CLI actions.
    'alias'      :  Add an alias: alias <alias_name> <actions>
                    Remove an alias: alias <alias_name>
                    Update an alias by creating a new alias with the same name.
                    Using 'alias' without parameters shows the current list of
                    aliases.
                    Aliases override existing actions and can contain
                    sequences of actions.
    'cd'         :  Change the working directory of the shell, e.g. 'cd ..'.
                    Note that on Windows, backslashes must be doubled, e.g.:
                    'cd C:\\'
    'check_for_update' : Check whether an update is available
    'docs'       :  Print a link to the online documentation.
    'exec'       :  Run a shell command, e.g.: 'exec ls -l'.
    'exit'       :  Exit the shell.
    'help'       :  Show this help message (available shell commands).
    'pop'        :  Restore saved active speaker state.
    'push'       :  Save the current active speaker, and unset the active
                    speaker.
    'rescan'     :  If your speaker doesn't appear in the 'speakers' list,
                    use this to perform a more comprehensive scan.
    'rescan_max' :  Try this if you're having having trouble finding all your
                    speakers.
    'set <spkr>' :  Set the active speaker using its name.
                    Use quotes when needed for the speaker name, e.g.,
                    'set "Front Reception"'. Unambiguous, partial,
                    case-insensitive matches are supported, e.g., 'set front'.
                    To unset the active speaker, omit the speaker name,
                    or just enter '0'.
    'sk'         :  Enters 'single keystroke' mode. (Also 'single-keystroke'.)
    'speakers'   :  List the names of all available speakers.
    'version'    :  Print the versions of SoCo-CLI, SoCo, and Python in use.
    
    The action syntax is the same as when using 'sonos' from the command line.
    If a speaker has been set in the shell, omit the speaker name from the
    action.

    Use the arrow keys for command history and command editing.
    
    [Not Available on Windows] Use the TAB key for autocompletion of shell
    commands, SoCo-CLI actions, aliases, and speaker names.
"""


def _get_speaker_names(use_local_speaker_list=False):
    if use_local_speaker_list:
        names = local_speaker_list().get_all_speaker_names()
    else:
        try:
            names = speaker_cache().get_all_speaker_names()
        except Exception as e:
            print(
                "Speaker listing failed: please check your network connection [{}]".format(
                    e
                )
            )
            names = []
    return names


def _print_speaker_list(use_local_speaker_list=False):
    print()
    names = _get_speaker_names(use_local_speaker_list=use_local_speaker_list)
    if len(names) > 0:
        names.insert(0, "Unset the active speaker")
        for index, name in enumerate(names, start=0):
            print("  ", str(index).rjust(2), ":", name)
        print()


def _save_readline_history():
    if RL:
        logging.info("Saving shell history")
        save_readline_history()


def _get_readline_history():
    if RL:
        logging.info("Reading shell history")
        get_readline_history()


def _restore_quotes(command):
    for index, parts in enumerate(command):
        if len(parts.split()) > 1:
            command[index] = '"' + parts + '"'


class AliasProcessor:

    # The arg substitution names %1, ..., %9
    _arg_names = tuple("%" + str(x) for x in range(1, 10))

    def __init__(self):
        self._used_aliases = []
        self._seq_number = 0
        self._recurse_level = 0
        self._command_count = 0
        self._index = 0
        self._command_list = []

    def process(self, command, am, command_list):

        self._recurse_level += 1

        logging.info(
            "Alias unpacking: recursion level {}, sequence number {}".format(
                self._recurse_level, self._seq_number
            )
        )

        alias_name = command[0]
        alias_parms = command[1:]

        self._command_list = command_list

        # Detect loops
        for used_alias in self._used_aliases:
            if used_alias[0] != self._recurse_level:
                if used_alias[1] == self._seq_number and used_alias[2] == alias_name:
                    # Alias name reused at different recursion levels but within
                    # the unpacking of the same sequence signifies a loop.
                    print("Error: Alias loop detected ... stopping".format(alias_name))
                    self._remove_added_commands()
                    return False
        self._used_aliases.append((self._recurse_level, self._seq_number, alias_name))

        alias_actions = am.action(alias_name)
        try:
            action_elements = shlex_split(alias_actions)
        except ValueError as error:
            print("Error: {}".format(error))
            return False

        cli_parser = CLIParser()
        cli_parser.parse(action_elements)
        sequences = cli_parser.get_sequences()

        logging.info("Unpacking the alias '{}' -> '{}'".format(alias_name, sequences))

        index = command_list.index()
        for idx, sequence in enumerate(sequences, start=1):
            self._seq_number = idx

            alias_parms_local = alias_parms.copy()

            # Deprecated
            if "_" in sequence:
                logging.info("Deprecated: Suppressing any additional parameters")
                print("*** Warning: The use of '_' is now deprecated in favour of")
                print("*** positional arguments (%1, %2, etc.).")
                position = sequence.index("_")
                sequence = sequence[:position]
            else:
                # Positional argument substitution: %1, %2, etc.
                alias_parms_used = []
                for i, item in enumerate(sequence):
                    if item in self._arg_names:
                        parm_index = int(item[1]) - 1
                        try:
                            sequence[i] = alias_parms_local[parm_index]
                            logging.info(
                                "Substituting '{}' for arg. {}".format(
                                    alias_parms_local[parm_index], item
                                )
                            )
                            # This allows reuse of the same parameter
                            alias_parms_used.append(parm_index)
                        except IndexError:
                            logging.info("No value found for arg. {}".format(item))
                            sequence[i] = None

                # Remove unsatisfied arguments and substituted parameters
                sequence = [x for x in sequence if x is not None]
                alias_parms_local = [
                    y
                    for x, y in enumerate(alias_parms_local)
                    if not x in alias_parms_used
                ]

            # Recurse if the sequence is itself an alias
            try:
                if sequence[0] in am.alias_names():
                    logging.info(
                        "Recursively unpacking the alias '{}'".format(sequence[0])
                    )
                    logging.info("Unpacking: '{}'".format(sequence + alias_parms_local))
                    if self.process(sequence + alias_parms_local, am, command_list):
                        index = command_list.index()
                    else:
                        return False

                # Not an alias, so insert the sequence in the command list
                # at the correct index, and increment the index
                else:
                    logging.info(
                        "Inserting new sequence {} at {}".format(sequence, index)
                    )
                    command_list.insert(index, sequence)
                    index += 1
                    self._command_count += 1
                    self._index = index

            except IndexError:
                logging.info("Empty sequence ... returning")
                return False

        self._recurse_level -= 1

        return True

    def _remove_added_commands(self):
        for _ in range(self._command_count):
            cmd = self._command_list.pop_next()
            logging.info("Removing command {}".format(cmd))


def _rescan(use_local_speaker_list=False, max_scan=False):
    try:
        if use_local_speaker_list:
            print("Using cached speaker list: no rescan performed")
        elif max_scan:
            logging.info("Full network rescan at max strength (timeout = 10.0s)")
            speaker_cache().scan(reset=True, scan_timeout_override=10.0)
            _print_speaker_list(use_local_speaker_list=use_local_speaker_list)
        else:
            logging.info("Full network rescan")
            speaker_cache().scan(reset=True)
            _print_speaker_list(use_local_speaker_list=use_local_speaker_list)
        _set_actions_and_commands_list(use_local_speaker_list=use_local_speaker_list)
    except Exception as e:
        print("Rescan failed: please check your network connection [{}]".format(e))


def _exec(command_line):
    """Runs a command as a subprocess, in its own shell.

    Args:
        command_line (list): The command to execute.
    """

    # Check for spaces in the script pathname
    if " " in command_line[0]:
        command_line[0] = '"' + command_line[0] + '"'

    # Convert command list to a unified command line
    command_line = " ".join(command_line)

    set_suspend_sighandling(suspend=True)
    try:
        logging.info("Running command: '{}'".format(command_line))
        subprocess.run(command_line, shell=True)
    except Exception as e:
        print(e)
    set_suspend_sighandling(suspend=False)


def _exec_action(speaker_ip, action, args):
    # Commands to run in a subprocess, to allow CTRL-C
    # to exit the subprocess only, and not the shell.
    if action not in ACTIONS_TO_EXEC_NO_SPEAKER:
        command_line = [sys.argv[0], speaker_ip, action, *args]
    else:  # Omit speaker
        command_line = [sys.argv[0], action, *args]

    for position, arg in enumerate(sys.argv[1:]):
        if arg.startswith("--log="):
            command_line.insert(1, arg)
            break
        if arg == "--log":
            command_line.insert(1, arg)
            command_line.append(sys.argv[1:][position + 1])
            break

    logging.info("Running '{}' in a subprocess. Terminate using CTRL-C.".format(action))
    _exec(command_line)



"""The main entry point into the 'sonos-discover' command."""

import argparse

from soco_cli.check_for_update import print_update_status
from soco_cli.speakers import Speakers
from soco_cli.utils import (
    check_args,
    configure_common_args,
    configure_logging,
    docs,
    error_report,
    logo,
    version,
)


def main():
    # Create the argument parser
    parser = argparse.ArgumentParser(
        prog="sonos-discover",
        usage="%(prog)s",
        description="Sonos speaker discovery utility",
    )
    parser.add_argument(
        "--print",
        "-p",
        action="store_true",
        default=False,
        help="Print the contents of the current speaker information file, and exit",
    )
    parser.add_argument(
        "--delete-local-speaker-cache",
        "-d",
        action="store_true",
        default=False,
        help="Delete the local speaker cache, if it exists",
    )
    parser.add_argument(
        "--subnets",
        type=str,
        help="Specify the networks or IP addresses to search, in dotted decimal/CIDR format",
    )
    # The rest of the optional args are common
    configure_common_args(parser)

    # Parse the command line
    args = parser.parse_args()

    configure_logging(args.log)

    if args.version:
        version()
        exit(0)

    if args.docs:
        docs()
        exit(0)

    if args.logo:
        logo()
        exit(0)

    if args.check_for_update:
        print_update_status()
        exit(0)

    # Create the Speakers object
    speaker_list = Speakers()

    if args.print:
        if speaker_list.load():
            speaker_list.print()
            exit(0)
        else:
            error_report("No current speaker data")

    if args.delete_local_speaker_cache:
        try:
            file = speaker_list.remove_save_file()
            print("Removed file: {}".format(file))
            exit(0)
        except Exception:
            error_report("No current speaker data file")

    # Parameter validation for various args
    message = check_args(args)
    if message:
        error_report(message)

    speaker_list._network_threads = args.network_discovery_threads
    speaker_list._network_timeout = args.network_discovery_timeout
    speaker_list._min_netmask = args.min_netmask
    if args.subnets is not None:
        speaker_list.subnets = args.subnets.split(",")

    try:
        speaker_list.discover()
        saved = speaker_list.save()
        speaker_list.print()
        if saved:
            print("Saved speaker data at: {}\n".format(speaker_list.save_pathname))
        else:
            print("No speakers discovered. No cache data saved or overwritten.")
    except Exception as e:
        error_report(str(e))


if __name__ == "__main__":
    # Catch all untrapped exceptions
    try:
        main()
        exit(0)
    except Exception as error:
        error_report(str(error))
        exit(1)



"""The main entry point into the 'sonos' command."""

import argparse
import logging
import pprint
import sys
import time
from os import environ as env
from signal import SIGINT, signal

import soco

from soco_cli.action_processor import list_actions
from soco_cli.aliases import AliasManager
from soco_cli.api import get_all_speakers, run_command
from soco_cli.check_for_update import print_update_status
from soco_cli.cmd_parser import CLIParser
from soco_cli.interactive import interactive_loop
from soco_cli.speakers import Speakers
from soco_cli.track_follow import track_follow
from soco_cli.utils import (
    RewindableList,
    check_args,
    configure_common_args,
    configure_logging,
    convert_to_seconds,
    create_speaker_cache,
    docs,
    error_report,
    get_speaker,
    logo,
    seconds_until,
    set_interactive,
    set_speaker_list,
    sig_handler,
    version,
)

from .wait_actions import process_wait

# Globals
pp = pprint.PrettyPrinter(width=100)


# Speaker name environment variable
ENV_SPKR = "SPKR"

# Local speaker cache environment variable
ENV_LOCAL = "USE_LOCAL_SPKR_CACHE"


def main():
    # Handle SIGINT
    signal(SIGINT, sig_handler)

    # Create the argument parser
    parser = argparse.ArgumentParser(
        prog="sonos",
        usage="%(prog)s <options> SPEAKER_NAME_OR_IP ACTION <parameters> < : ...>",
        description="Command line utility for controlling Sonos speakers",
    )
    # A variable number of arguments depending on the action
    parser.add_argument(
        "parameters", nargs="*", help="Sequences of SPEAKER ACTION <parameters> : ..."
    )
    # Optional arguments
    parser.add_argument(
        "--use-local-speaker-list",
        "-l",
        action="store_true",
        default=False,
        help="Use the local speaker list instead of SoCo discovery",
    )
    parser.add_argument(
        "--refresh-local-speaker-list",
        "-r",
        action="store_true",
        default=False,
        help="Refresh the local speaker list",
    )
    parser.add_argument(
        "--actions",
        action="store_true",
        default=False,
        help="Print the list of available actions",
    )
    parser.add_argument(
        "--commands",
        action="store_true",
        default=False,
        help="Print the list of available actions",
    )
    parser.add_argument(
        "--interactive",
        "-i",
        action="store_true",
        default=False,
        help="Enter interactive mode",
    )
    parser.add_argument(
        "--no-env",
        action="store_true",
        default=False,
        help="Ignore the 'SPKR' environment variable, if set",
    )
    parser.add_argument(
        "--sk",
        action="store_true",
        default=False,
        help="Enter single keystroke mode in the interactive shell",
    )
    parser.add_argument(
        "--save_aliases",
        type=str,
        help="Save the current shell aliases to the supplied filename and exit",
    )
    parser.add_argument(
        "--load_aliases",
        type=str,
        help="Load shell aliases from the supplied filename and exit (aliases are merged)",
    )
    parser.add_argument(
        "--overwrite_aliases",
        type=str,
        help="Overwrite current shell aliases with those from the supplied filename and exit",
    )
    # The rest of the optional args are common
    configure_common_args(parser)

    # Parse the command line
    args = parser.parse_args()

    configure_logging(args.log)

    if args.version:
        version()
        exit(0)

    if args.docs:
        docs()
        exit(0)

    if args.logo:
        logo()
        exit(0)

    if args.check_for_update:
        print_update_status()
        exit(0)

    if args.actions or args.commands:
        list_actions()
        exit(0)

    if args.save_aliases:
        am = AliasManager()
        am.load_aliases()
        if am.save_aliases_to_file(args.save_aliases):
            print("Saved shell aliases to '{}'".format(args.save_aliases), flush=True)
        else:
            print(
                "Failed to save shell aliases to '{}'".format(args.save_aliases),
                flush=True,
            )
        exit(0)

    if args.load_aliases:
        am = AliasManager()
        am.load_aliases()
        if am.load_aliases_from_file(args.load_aliases):
            print(
                "Loaded and merged shell aliases from '{}'".format(args.load_aliases),
                flush=True,
            )
        else:
            print(
                "Failed to load shell aliases from '{}'".format(args.load_aliases),
                flush=True,
            )
        exit(0)

    if args.overwrite_aliases:
        am = AliasManager()
        if am.load_aliases_from_file(args.overwrite_aliases):
            print(
                "Loaded and saved shell aliases from '{}'".format(
                    args.overwrite_aliases
                ),
                flush=True,
            )
        else:
            print(
                "Failed to load shell aliases from '{}'".format(args.overwrite_aliases),
                flush=True,
            )
        exit(0)

    if len(args.parameters) == 0 and not args.interactive:
        print("No parameters. Use 'sonos --help' for usage information", flush=True)
        exit(0)

    message = check_args(args)
    if message:
        error_report(message)

    use_local_speaker_list = args.use_local_speaker_list
    if env.get(ENV_LOCAL) == "TRUE" and not args.no_env:
        logging.info(
            "Env. var. '{}' set to 'TRUE ... using local speaker list".format(ENV_LOCAL)
        )
        use_local_speaker_list = True
    if use_local_speaker_list:
        speaker_list = Speakers(
            network_threads=args.network_discovery_threads,
            network_timeout=args.network_discovery_timeout,
            min_netmask=args.min_netmask,
        )
        if args.refresh_local_speaker_list or not speaker_list.load():
            logging.info("Start speaker discovery")
            speaker_list.discover()
            speaker_list.save()
        set_speaker_list(speaker_list)
    else:
        # Create the local speaker cache in the utils module
        create_speaker_cache(
            max_threads=args.network_discovery_threads,
            scan_timeout=args.network_discovery_timeout,
            min_netmask=args.min_netmask,
        )

    # Is $SPKR set in the environment?
    env_speaker = None
    if not args.no_env:
        env_speaker = env.get(ENV_SPKR)
        if env_speaker:
            logging.info("Found 'SPKR' environment variable: '{}'".format(env_speaker))
        else:
            logging.info("No 'SPKR' environment variable set")

    if args.interactive:
        sk = bool(args.sk)
        speaker_name = None
        if len(args.parameters):
            speaker_name = args.parameters[0]
        interactive_loop(
            speaker_name,
            use_local_speaker_list=use_local_speaker_list,
            no_env=args.no_env,
            single_keystroke=sk,
        )
        exit(0)

    cli_parser = CLIParser()
    cli_parser.parse(args.parameters)
    sequences = cli_parser.get_sequences()

    cumulative_exit_code = 0

    # Loop through processing command sequences
    logging.info("Found {} action sequence(s): {}".format(len(sequences), sequences))
    rewindable_sequences = RewindableList(sequences)
    loop_iterator = None
    sequence_pointer = 0

    # There is a notional 'loop' action before the first command sequence
    loop_pointer = -1

    loop_start_time = None
    for sequence in rewindable_sequences:
        try:
            speaker_name = sequence[0]

            # Special case: the 'loop_to_start' action
            if speaker_name.lower() == "loop_to_start":
                if len(sequence) != 1:
                    error_report("Action 'loop_to_start' takes no parameters")
                # Reset pointers, rewind and continue
                loop_pointer = -1
                sequence_pointer = 0
                logging.info("Rewind to start of command sequences")
                rewindable_sequences.rewind()
                continue

            # Special case: the 'loop' action
            if speaker_name.lower() == "loop":
                if len(sequence) == 2:
                    if loop_iterator is None:
                        try:
                            loop_iterator = int(sequence[1])
                            if loop_iterator <= 0:
                                raise ValueError
                            logging.info(
                                "Looping for {} iteration(s)".format(loop_iterator)
                            )
                        except ValueError:
                            error_report(
                                "Action 'loop' takes no parameters, or a number of iterations (> 0)"
                            )
                    loop_iterator -= 1
                    logging.info("Loop iterator countdown = {}".format(loop_iterator))
                    if loop_iterator <= 0:
                        # Reset variables, stop iteration and continue
                        loop_iterator = None
                        loop_pointer = sequence_pointer
                        sequence_pointer += 1
                        continue
                logging.info("Rewinding to command number {}".format(loop_pointer + 2))
                rewindable_sequences.rewind_to(loop_pointer + 1)
                sequence_pointer = loop_pointer + 1
                continue

            # Special case: the 'loop_for' action
            if speaker_name.lower() == "loop_for":
                if len(sequence) != 2:
                    error_report(
                        "Action 'loop_for' requires one parameter (check spaces around the ':' separator)"
                    )
                loop_duration = 0
                if loop_start_time is None:
                    loop_start_time = time.time()
                    try:
                        loop_duration = convert_to_seconds(sequence[1])
                    except ValueError:
                        error_report(
                            "Action 'loop_for' requires one parameter (duration >= 0)"
                        )
                    logging.info(
                        "Starting action 'loop_for' for duration {}s".format(
                            loop_duration
                        )
                    )
                else:
                    if time.time() - loop_start_time >= loop_duration:
                        logging.info(
                            "Ending action 'loop_for' after duration {}s".format(
                                loop_duration
                            )
                        )
                        loop_start_time = None
                        continue
                logging.info("Rewinding to command number {}".format(loop_pointer + 2))
                rewindable_sequences.rewind_to(loop_pointer + 1)
                sequence_pointer = loop_pointer
                continue

            # Special case: the 'loop_until' action
            if speaker_name.lower() == "loop_until":
                loop_duration = 0
                if len(sequence) != 2:
                    error_report(
                        "Action 'loop_until' requires one parameter (check spaces around the ':' separator)"
                    )
                if loop_start_time is None:
                    loop_start_time = time.time()
                    try:
                        loop_duration = seconds_until(sequence[1])
                    except:
                        error_report(
                            "Action 'loop_until' requires one parameter (stop time)"
                        )
                    logging.info(
                        "Starting action 'loop_until' for duration {}s".format(
                            loop_duration
                        )
                    )
                else:
                    if time.time() - loop_start_time >= loop_duration:
                        logging.info(
                            "Ending action 'loop_until' after duration {}s".format(
                                loop_duration
                            )
                        )
                        loop_start_time = None
                        continue
                logging.info("Rewinding to command number {}".format(loop_pointer + 2))
                rewindable_sequences.rewind_to(loop_pointer + 1)
                sequence_pointer = loop_pointer
                continue

            # Special case: the 'wait' actions
            if speaker_name in ["wait", "wait_for", "wait_until"]:
                process_wait(sequence)
                continue

            # Use the speaker name from the environment?
            if env_speaker:
                logging.info(
                    "Getting speaker name '{}' from the $SPKR environment variable".format(
                        env_speaker
                    )
                )
                sequence.insert(0, env_speaker)
                speaker_name = env_speaker

            # General action processing
            if len(sequence) < 2:
                error_report(
                    "At least 2 parameters required in action sequence '{}'".format(
                        sequence
                    )
                )
            action = sequence[1].lower()
            args = sequence[2:]
            if speaker_name.lower() == "_all_":
                if use_local_speaker_list:
                    speakers = speaker_list.get_all_speakers()
                else:
                    speakers = get_all_speakers(use_scan=True)
                logging.info(
                    "Performing action '{}' on all visible, coordinator speakers".format(
                        action
                    )
                )
                for speaker in speakers:
                    if speaker.is_visible and speaker.is_coordinator:
                        logging.info(
                            "Performing action '{}' on speaker '{}'".format(
                                action, speaker.player_name
                            )
                        )
                        print(speaker.player_name + ": ", end="", flush=True)
                        exit_code, output_msg, error_msg = run_command(
                            speaker,
                            action,
                            *args,
                            use_local_speaker_list=use_local_speaker_list,
                        )
                        if exit_code == 0:
                            if len(output_msg) != 0:
                                print(output_msg, flush=True)
                            else:
                                print("OK", flush=True)
                        elif len(error_msg) != 0:
                            print(error_msg, file=sys.stderr, flush=True)
                        cumulative_exit_code += exit_code
            else:
                speaker = get_speaker(speaker_name, use_local_speaker_list)
                if not speaker:
                    print(
                        "Error: Speaker '{}' not found".format(speaker_name),
                        file=sys.stderr,
                        flush=True,
                    )
                    cumulative_exit_code += 1
                else:
                    # Special case of 'track_follow' action
                    if action in ["track_follow", "tf", "track_follow_compact", "tfc"]:
                        if len(args) > 0:
                            print(
                                "Error: Action '{}' takes no parameters".format(action),
                                file=sys.stderr,
                                flush=True,
                            )
                            continue
                        # Does not return
                        compact = action in ["track_follow_compact", "tfc"]
                        track_follow(
                            speaker,
                            use_local_speaker_list=use_local_speaker_list,
                            break_on_pause=False,
                            compact=compact,
                        )
                    # Standard action processing
                    exit_code, output_msg, error_msg = run_command(
                        speaker,
                        action,
                        *args,
                        use_local_speaker_list=use_local_speaker_list,
                    )
                    if exit_code == 0 and len(output_msg) != 0:
                        print(output_msg, flush=True)
                    elif len(error_msg) != 0:
                        print(error_msg, file=sys.stderr, flush=True)
                    cumulative_exit_code += exit_code

        except Exception as e:
            print("Error: ", str(e), flush=True)
            cumulative_exit_code += 1

        sequence_pointer += 1

    exit(cumulative_exit_code)


if __name__ == "__main__":
    # Catch all untrapped exceptions
    try:
        main()
        exit(0)
    except Exception as error:
        error_report(str(error))
        exit(1)



"""Process the speaker-independent 'wait' actions."""

import logging
import time

from soco_cli.utils import convert_to_seconds, error_report, seconds_until


def process_wait(sequence):
    if sequence[0] in ["wait", "wait_for"]:
        duration = 0
        if len(sequence) != 2:
            error_report(
                "Action 'wait' requires 1 parameter (check spaces around the ':' separator?)"
            )
            return
        action = sequence[1].lower()
        try:
            duration = convert_to_seconds(action)
        except ValueError:
            error_report(
                "Action 'wait' requires positive number of hours, seconds or minutes + 'h/m/s', or HH:MM(:SS)"
            )
        logging.info("Waiting for {}s".format(duration))
        time.sleep(duration)

    # Special case: the 'wait_until' action
    elif sequence[0] in ["wait_until"]:
        if len(sequence) != 2:
            error_report(
                "'wait_until' requires 1 parameter (check spaces around the ':' separator?)"
            )
            return
        try:
            action = sequence[1].lower()
            duration = seconds_until(action)
            logging.info("Waiting for {}s".format(duration))
            time.sleep(duration)
        except ValueError:
            error_report(
                "'wait_until' requires parameter: time in 24hr HH:MM(:SS) format"
            )



"""Checks GitHub for a later version of SoCo-CLI"""

import logging
from urllib.request import urlopen

from soco_cli.__init__ import __version__
from soco_cli.utils import error_report

init_file_url = (
    "https://raw.githubusercontent.com/avantrec/soco-cli/master/soco_cli/__init__.py"
)


def get_latest_version():
    try:
        file = urlopen(init_file_url, timeout=3.0)
    except Exception as e:
        error_report(
            "Unable to get latest version information from GitHub: {}".format(e)
        )
        return False

    for line in file:
        decoded_line = line.decode("utf-8")
        if "__version__" in decoded_line:
            latest_version = (
                decoded_line.replace("__version__ = ", "")
                .replace('"', "")
                .replace("\n", "")
            )
            logging.info("Latest version is v{}".format(latest_version))
            break
    else:
        logging.info("Unable to find latest version")
        return None

    return latest_version


def print_update_status():
    latest_version = get_latest_version()
    if latest_version is not None:
        if __version__ == latest_version:
            print("SoCo-CLI is up to date")
        else:
            print("An update is available: v" + latest_version)
        return True

    return False


def update_available():
    if __version__ == get_latest_version():
        return False

    return True



import logging
import re

from datetime import datetime, timezone
from time import sleep
from soco_cli.api import run_command


def track_follow(
    speaker, use_local_speaker_list=False, break_on_pause=True, compact=False
):
    """Print out the 'track' details each time the track changes.

    Args:
        speaker (SoCo): The speaker to follow.
        use_local_speaker_list (bool, optional): Use cached discovery.
        break_on_pause (bool, optional): Whether to return control if the
            speaker enters the paused or stopped playback states.

    This function operates as if 'outside' the main program logic, because
    it needs to output intermediate results as it executes. Hence, the
    'run_command()' API call is used.
    """

    def timestamp(short=False):
        local_tz = datetime.now(timezone.utc).astimezone().tzinfo
        if not short:
            return datetime.now(tz=local_tz).strftime("%H:%M:%S (%Z) %d-%b-%Y")
        else:
            return datetime.now(tz=local_tz).strftime("%H:%M")

    counter = 1
    print(flush=True)
    while True:
        # If stopped, wait for the speaker to start playback
        _, state, _ = run_command(
            speaker, "state", use_local_speaker_list=use_local_speaker_list
        )
        if state in [
            "STOPPED",
            "PAUSED_PLAYBACK",
        ]:
            if not compact:
                print(
                    " [{}] Playback is stopped or paused at {}\n".format(
                        speaker.player_name, timestamp()
                    ),
                    flush=True,
                )
            else:
                print(
                    "{:5d}: [{}] Playback is stopped or paused".format(
                        counter, timestamp(short=True)
                    )
                )
                counter += 1
            if break_on_pause:
                logging.info("Playback is paused/stopped; returning")
                break
            logging.info("Playback is paused/stopped; waiting for start")
            run_command(
                speaker, "wait_start", use_local_speaker_list=use_local_speaker_list
            )
            logging.info("Speaker has started playback")

        # Print the track info
        exit_code, output, error_msg = run_command(
            speaker, "track", use_local_speaker_list=use_local_speaker_list
        )
        if exit_code == 0:
            # Manipulate output
            if "Using Line In" in output:
                line_in = True
                output = "   Playing from Line In\n"
            else:
                line_in = False
                output = output.split("\n ", 1)[1]
            if not compact:
                # Remove some of the line entries
                output = re.sub(".*Playback.*\\n", "", output)
                # output = re.sub(".*Elapsed.*\\n", "", output)
                output = re.sub(".*URI.*\\n", "", output)
                output = re.sub(".*Uri.*\\n", "", output)
                # Prefix speaker name and timestamp
                output = (
                    " [{}] Playing at ".format(speaker.player_name)
                    + timestamp()
                    + ":\n"
                    + output
                )
            else:  # Compact (one line) output
                if line_in:
                    output = "{:5d}: [{}] Playing from Line In".format(
                        counter, timestamp(short=True)
                    )
                else:
                    # Ordering of keys determines output order
                    keys = [
                        "Channel:",
                        "Radio Show:",
                        "Artist:",
                        "Creator(s):",
                        "Book Title:",
                        "Chapter:",
                        "Album:",
                        "Podcast:",
                        "Title:",
                        "Episode:",
                        "Release Date:",
                        "Narrator(s):",
                    ]
                    elements = {}
                    for line in output.splitlines():
                        for key in keys:
                            if key in line:
                                elements[key] = line.replace(key, "").lstrip()
                    output = "{:5d}: [{}] ".format(counter, timestamp(short=True))

                    # Prune fields for audio books
                    if "Book Title:" in elements:
                        elements.pop("Title:", None)
                        elements.pop("Narrator(s):", None)
                    first = True
                    for key in keys:
                        value = elements.pop(key, None)
                        if value:
                            if not first:
                                output = output + "| "
                            else:
                                first = False
                            output = output + key + " " + value + " "

            print(output, flush=True)
        else:
            error_out = "{:5d}: [{}] {}".format(
                counter, timestamp(short=True), error_msg
            )
            print(error_out, flush=True)

        logging.info("Waiting for end of track")
        run_command(
            speaker, "wait_end_track", use_local_speaker_list=use_local_speaker_list
        )

        logging.info("Waiting 1.5s for playback to stabilise")
        sleep(1.5)
        counter += 1



import io
import re

import setuptools

with open("PYPI_README.md", "r") as fh:
    long_description = fh.read()

src = io.open("soco_cli/__init__.py", encoding="utf-8").read()
metadata = dict(re.findall('__([a-z]+)__ = "([^"]+)"', src))
# docstrings = re.findall('"""(.*?)"""', src, re.MULTILINE | re.DOTALL)

REQUIREMENTS = list(open("requirements.txt"))
VERSION = metadata["version"]

setuptools.setup(
    name="soco-cli",
    version=VERSION,
    author="Avantrec Ltd",
    author_email="soco_cli@avantrec.com",
    description="Sonos command line control utility, based on SoCo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avantrec/soco-cli",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    python_requires=">=3.5",
    install_requires=REQUIREMENTS,
    entry_points={
        "console_scripts": [
            "sonos=soco_cli.sonos:main",
            "soco=soco_cli.sonos:main",
            "sonos-discover=soco_cli.sonos_discover:main",
            "soco-discover=soco_cli.sonos_discover:main",
            "sonos-http-api-server=soco_cli.http_api:main",
            "soco-http-api-server=soco_cli.http_api:main",
        ]
    },
)
